/* NAME:
 *	unfsmnt - [u]mount routines for BSD and SunOS
 *
 * SYNOPSIS:
 *	mount_unfs()
 *	umount_unfs()
 *
 * DESCRIPTION:
 *	These routines handle the MOUNT and UMOUNT RPC's
 *	The mount logic is based on NetBSD's mount_nfs.
 *
 * AUTHOR:
 *	Simon J. Gerraty <sjg@zen.void.oz.au>
 */
/*
 *      @(#)Copyright (c) 1994, Simon J. Gerraty.
 *      
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@zen.void.oz.au
 */

#ifndef lint
static char RCSid[] = "$Id: unfsmnt.c,v 1.6 1995/05/13 07:22:33 sjg Exp $";
#endif
#include <stdio.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/stat.h>
#include <sys/syslog.h>
#include <sys/types.h>
#include <errno.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>

#include <rpc/rpc.h>

#include "mount_prot.h"
#include "nfs_prot.h"

#include "unfsd.h"

/*
 * Sadly, we either define these ourselves, or we include nfs.h and a
 * bunch of others under BSD.  Including nfs.h etc means we cannot
 * include nfs_prot.h, which came from rpcgen.
 * These values come from the BSD headers indicated.
 */
/*
 * nfs/nfs.h
 */
#ifndef NFS_WSIZE
# define NFS_WSIZE       8192
#endif
#ifndef NFS_RSIZE
# define NFS_RSIZE       8192
#endif
#ifndef NFS_HZ
# define NFS_HZ          25              /* ticks/s for NFS timeouts */
#endif
#ifndef NFS_TIMEO
# define NFS_TIMEO	(1*NFS_HZ)
#endif
#ifndef NFS_RETRANS
# define NFS_RETRANS	10
#endif
#ifndef NFS_MAXGRPS
# define NFS_MAXGRPS	16
#endif
#ifndef NFS_DEFRAHEAD
# define NFS_DEFRAHEAD	1
#endif
/*
 * nfs/nqnfs.h
 */
#ifndef NQ_DEFLEASE
# define NQ_DEFLEASE	30
#endif
#ifndef NQ_DEADTHRESH
# define NQ_DEADTHRESH	9
#endif

#ifndef M_NEWTYPE
# define M_NEWTYPE 0
#endif

/*
 * Some of this is lifted from NetBSD's mount_nfs.c
 */
struct nfs_args nfsdefargs = {
#ifdef BSD4_4
	(struct sockaddr *)0,
	sizeof (struct sockaddr_in),
	SOCK_DGRAM,
	0,
	(nfsv2fh_t *)0,
	0,
	NFS_WSIZE,
	NFS_RSIZE,
	NFS_TIMEO,
	NFS_RETRANS,
	NFS_MAXGRPS,
	NFS_DEFRAHEAD,
	NQ_DEFLEASE,
	NQ_DEADTHRESH,
	(char *)0,
#else  /* SunOS compatible? */
	(struct sockaddr_in *)0,
	(caddr_t)0,
	0,
	NFS_WSIZE,
	NFS_RSIZE,
	NFS_TIMEO,
	NFS_RETRANS,
	(char *)0,
	0,
	0,
	0,
	0,
	(char *)0,
#if !defined(sun386)
	(struct pathcnf *)0,
#endif
#endif
};

static int
bust_spec(mntspec, port, sa, hostp, hostdirp)
	const char *mntspec;
	int port;
	struct sockaddr_in *sa;
	char **hostp, **hostdirp;
{
	static char spec[256];
        struct hostent          *hp;
	register char *rcp;

	strncpy(spec, mntspec, sizeof (spec) - 1);
	spec[sizeof (spec) - 1] = '\0';

	if ((rcp = strchr(spec, ':')) != NULL) {
		*hostp = spec;
		*rcp++ = '\0';
		*hostdirp = rcp;
		if ((rcp = strchr(rcp, ':')) != NULL) {
			port = atoi(*hostdirp);
			*hostdirp = rcp + 1;
		}
	} else {
		warnx("no <host>:[port:]<dirpath> spec in %s", mntspec);
		return 0;
	}
	memset((char *)sa, 0, sizeof (struct sockaddr_in));
	sa->sin_family = AF_INET;
	
	if (isdigit(*hostp[0])) {
		if ((sa->sin_addr.s_addr = inet_addr(*hostp)) < 0) {
			warnx("bad net address %s", *hostp);
			return 0;
		}
	} else if ((hp = gethostbyname(*hostp)) != NULL) {
#ifndef BSD4_4
		memcpy((char *) &sa->sin_addr, hp->h_addr, hp->h_length);
#else
		memmove((char *) &sa->sin_addr, hp->h_addr, hp->h_length);
#endif
	} else {
		warnx("unknown server %s", *hostp);
		return 0;
	}
	if ((UnfsFlags & (UNFS_UNMOUNT|UNFS_CLNT)) ==
	    (UNFS_UNMOUNT|UNFS_CLNT)) {
		/*
		 * Unmounting a TCP session.
		 * If kernel only supports UDP or we are using AUTH
		 * We will have a local unfsc process running.
		 * So look up localhost.localport, otherwise its server.port
		 */
		port = uclnts_get(UnfsMntSpec, UnfsMntPt);
		sa->sin_addr.s_addr = htonl(LOCALHOST_ADDR);
	}
	sa->sin_port = htons(port);
	if (UnfsFlags & UNFS_DEBUG)
		(void) fprintf(stderr, "%s -> %s\n", UnfsMntSpec,
			       ip_caller_id(sa, 0));
	return 1;
}

static CLIENT *
rpc_setup(s, sa)
	int *s;
	struct sockaddr_in *sa;
{
	static CLIENT *clp;
	struct timeval pertry;

	if (UnfsFlags & UNFS_TCP) {
		clp = clnttcp_create(sa,
				     MOUNTPROG,
				     MOUNTVERS,
				     s, 0, 0);
	} else {
		pertry.tv_sec = 3;
		pertry.tv_usec = 0;
		clp = clntudp_create(sa,
				     MOUNTPROG,
				     MOUNTVERS,
				     pertry, s);
	}
	if (clp == NULL) {
		if (!(UnfsFlags & UNFS_ISBGRND))
			clnt_pcreateerror("Cannot MNT PRC");
	} else {
		clp->cl_auth = authunix_create_default();
	}
	return clp;
}
	 
int
mount_unfs(mntflags)
	int mntflags;
{
	struct sockaddr_in sa;
	struct timeval try;
	enum clnt_stat clnt_stat;
	CLIENT *clp;
	char *hostp, *hostdirp;
	int s = RPC_ANYSOCK;
	int tries = 1000;
	fhstatus	nfhret;
	struct nfs_args nfsargs, *nfsargsp;
	
	if (!bust_spec(UnfsMntSpec, UnfsPort, &sa, &hostp, &hostdirp))
		return 1;

	UnfsMntedDir = hostdirp;
	nfsargs = nfsdefargs;
	nfsargsp = &nfsargs;

#ifndef NO_FWTK
	if ((UnfsFlags & (UNFS_CLNT|UNFS_AUTH)) == (UNFS_CLNT|UNFS_AUTH)) {
		/*
		 * ok we need strong authentication,
		 * but it only makes sense if we are going to
		 * run unfsclnt().
		 */
		if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0 ||
		    bindresvport(s, (struct sockaddr_in *) 0) < 0 ||
		    connect(s, (struct sockaddr *) &sa,
			    sizeof(struct sockaddr_in)) < 0 ||
		    get_auth(s) <= 0) {
			warnx("Authorization failed for %s",
			      *hostp);
			return 1;
		}
	}
#endif
	
	nfhret.fhs_status = EACCES;
		
	while (tries > 0) {
		if ((clp = rpc_setup(&s, &sa)) != NULL) {
			
			try.tv_sec = 10;
			try.tv_usec = 0;
			clnt_stat = clnt_call(clp, MOUNTPROC_MNT,
					      xdr_dirpath,
					      &hostdirp,
					      xdr_fhstatus,
					      &nfhret,
					      try);
				
			if (clnt_stat != RPC_SUCCESS) {
				if (!(UnfsFlags & UNFS_ISBGRND))
					warnx("%s", clnt_sperror(clp,
								 "bad MNT RPC"));
			} else {
				if (!(UnfsFlags & UNFS_CLNT)) {
					/* finished with these... */
					auth_destroy(clp->cl_auth);
					clnt_destroy(clp);
				}
				tries = 0;
			}
		}
		if (tries-- > 0) {
			if (UnfsFlags & UNFS_BGRND) {
				UnfsFlags &= ~UNFS_BGRND;
#ifndef HAVE_SETSID
				setpgrp();
#endif
				switch (fork()) {
				case 0:
#ifdef HAVE_SETSID
					(void) setsid();
#endif
					(void) close(STDIN_FILENO);
					(void) close(STDOUT_FILENO);
					(void) close(STDERR_FILENO);
					(void) chdir("/");
					break;
				case -1:
					err(1, "cannot fork");
				default:
					exit(0);
				}
				UnfsFlags |= UNFS_ISBGRND;
			}
			sleep(60);
		}
	}
	/*
	 * did we succeed?
	 */
	if (nfhret.fhs_status) {
		if (UnfsFlags & UNFS_ISBGRND)
			exit(1);
		
		errno = nfhret.fhs_status;
		warn("can't access %s", UnfsMntSpec);
		return 1;
	}
	if (UnfsFlags & UNFS_CLNT) {
		/* this call resets sa */
		fork_unfsc(s, clp, &sa,
			      UnfsMntPt, UnfsMntSpec);
		/* parent is finished with these... */
		auth_destroy(clp->cl_auth);
		clnt_destroy(clp);
		if (UnfsTimeo < 0)
			UnfsTimeo = 30;		/* 3 secs */
	}
	if (UnfsFlags & UNFS_INTR)
		nfsargsp->flags |= NFSMNT_INT;
	if (UnfsFlags & UNFS_SOFT)
		nfsargsp->flags |= NFSMNT_SOFT;
	if (UnfsTimeo > 0) {
		nfsargsp->timeo = UnfsTimeo;
		nfsargsp->flags |= NFSMNT_TIMEO;
	}
	if (UnfsRsize > 0) {
		nfsargsp->rsize = UnfsRsize;
		nfsargsp->flags |= NFSMNT_RSIZE;
	}
	if (UnfsWsize > 0) {
		nfsargsp->rsize = UnfsWsize;
		nfsargsp->flags |= NFSMNT_WSIZE;
	}
	
#ifdef BSD4_4
	nfsargsp->addr = (struct sockaddr *) &sa;
	nfsargsp->addrlen = sizeof (sa);
#else
	nfsargsp->addr = (struct sockaddr_in *) &sa;
#endif
#ifdef NFSMNT_HOSTNAME
	nfsargsp->flags |= NFSMNT_HOSTNAME;
#endif
#ifdef BSD4_4
	nfsargsp->fh = (nfsv2fh_t  *) &nfhret.fhstatus_u;
	nfsargsp->hostname = UnfsMntSpec;
#else
	nfsargsp->fh = (caddr_t) &nfhret.fhstatus_u;
	nfsargsp->hostname = hostp;
	if ((nfsargsp->netname = strchr(hostp, '.')) != NULL)
		*nfsargsp->netname++ = '\0';
#endif
	if (mount("nfs", UnfsMntPt, M_NEWTYPE|mntflags, nfsargsp))
		err(1, "%s", UnfsMntPt);
	return 0;
}


int
umount_unfs()
{
	struct sockaddr_in sa;
	enum clnt_stat clnt_stat;
	struct timeval try;
	CLIENT *clp;
	char *hostp, *hostdirp;
	int s = RPC_ANYSOCK;
	int sts = 1;
	
	if (bust_spec(UnfsMntSpec, UnfsPort, &sa, &hostp, &hostdirp)) {
		/*
		 * get it out of the kernel if we can
		 */
		(void) unmount(UnfsMntPt, 0);

#ifdef KERNEL_TCP_NFS
		if ((UnfsFlags & (UNFS_TCP|UNFS_AUTH)) == UNFS_TCP)
			return 0;		/* skip RPC */
#endif
		UnfsFlags &= ~UNFS_TCP;
		
		if ((clp = rpc_setup(&s, &sa)) != NULL) {
			try.tv_sec = 20;
			try.tv_usec = 0;
			clnt_stat = clnt_call(clp,
					      MOUNTPROC_UMNT,
					      xdr_dirpath, &hostdirp,
					      xdr_void, (caddr_t)0,
					      try);
			if (clnt_stat != RPC_SUCCESS) {
				clnt_perror(clp, "Bad MNT RPC");
			} else {
				auth_destroy(clp->cl_auth);
				clnt_destroy(clp);
				sts = 0;
			}
		}
	}
	return sts;
}
